@@ -7,7 +7,12 @@ void Unpack::Unpack5(bool Solid)
     UnpInitData(Solid);
     if (!UnpReadBuf())
       return;
-    if (!ReadBlockHeader(Inp,BlockHeader) || !ReadTables(Inp,BlockHeader,BlockTables))
+
+    // Check TablesRead5 to be sure that we read tables at least once
+    // regardless of current block header TablePresent flag.
+    // So we can safefly use these tables below.
+    if (!ReadBlockHeader(Inp,BlockHeader) ||
+        !ReadTables(Inp,BlockHeader,BlockTables) || !TablesRead5)
       return;
   }
 
@@ -37,7 +42,7 @@ void Unpack::Unpack5(bool Solid)
         break;
     }
 
-    if (((WriteBorder-UnpPtr) & MaxWinMask)<MAX_LZ_MATCH+3 && WriteBorder!=UnpPtr)
+    if (((WriteBorder-UnpPtr) & MaxWinMask)<MAX_INC_LZ_MATCH && WriteBorder!=UnpPtr)
     {
       UnpWriteBuf();
       if (WrittenFileSize>DestUnpSize)
@@ -173,6 +178,8 @@ bool Unpack::ReadFilter(BitInput &Inp,UnpackFilter &Filter)
 
   Filter.BlockStart=ReadFilterData(Inp);
   Filter.BlockLength=ReadFilterData(Inp);
+  if (Filter.BlockLength>MAX_FILTER_BLOCK_SIZE)
+    Filter.BlockLength=0;
 
   Filter.Type=Inp.fgetbits()>>13;
   Inp.faddbits(3);
@@ -189,11 +196,15 @@ bool Unpack::ReadFilter(BitInput &Inp,UnpackFilter &Filter)
 
 bool Unpack::AddFilter(UnpackFilter &Filter)
 {
-  if (Filters.Size()>=MAX_UNPACK_FILTERS-1)
+  if (Filters.Size()>=MAX_UNPACK_FILTERS)
+  {
     UnpWriteBuf(); // Write data, apply and flush filters.
+    if (Filters.Size()>=MAX_UNPACK_FILTERS)
+      InitFilters(); // Still too many filters, prevent excessive memory use.
+  }
 
   // If distance to filter start is that large that due to circular dictionary
-  // mode it points to old not written yet data, then we set 'NextWindow'
+  // mode now it points to old not written yet data, then we set 'NextWindow'
   // flag and process this filter only after processing that older data.
   Filter.NextWindow=WrPtr!=UnpPtr && ((WrPtr-UnpPtr)&MaxWinMask)<=Filter.BlockStart;
 
@@ -250,7 +261,7 @@ void Unpack::UnpWriteBuf()
   for (size_t I=0;I<Filters.Size();I++)
   {
     // Here we apply filters to data which we need to write.
-    // We always copy data to virtual machine memory before processing.
+    // We always copy data to another memory block before processing.
     // We cannot process them just in place in Window buffer, because
     // these data can be used for future string matches, so we must
     // preserve them in original form.
@@ -261,7 +272,7 @@ void Unpack::UnpWriteBuf()
     if (flt->NextWindow)
     {
       // Here we skip filters which have block start in current data range
-      // due to address warp around in circular dictionary, but actually
+      // due to address wrap around in circular dictionary, but actually
       // belong to next dictionary block. If such filter start position
       // is included to current write range, then we reset 'NextWindow' flag.
       // In fact we can reset it even without such check, because current
@@ -288,7 +299,7 @@ void Unpack::UnpWriteBuf()
       }
       if (BlockLength<=WriteSizeLeft)
       {
-        if (BlockLength>0)
+        if (BlockLength>0) // We set it to 0 also for invalid filters.
         {
           uint BlockEnd=(BlockStart+BlockLength)&MaxWinMask;
 
@@ -372,7 +383,8 @@ void Unpack::UnpWriteBuf()
   }
 
   // We prefer to write data in blocks not exceeding UNPACK_MAX_WRITE
-  // instead of potentially huge MaxWinSize blocks.
+  // instead of potentially huge MaxWinSize blocks. It also allows us
+  // to keep the size of Filters array reasonable.
   WriteBorder=(UnpPtr+Min(MaxWinSize,UNPACK_MAX_WRITE))&MaxWinMask;
 
   // Choose the nearest among WriteBorder and WrPtr actual written border.
@@ -393,9 +405,11 @@ byte* Unpack::ApplyFilter(byte *Data,uint DataSize,UnpackFilter *Flt)
       {
         uint FileOffset=(uint)WrittenFileSize;
 
-        const int FileSize=0x1000000;
+        const uint FileSize=0x1000000;
         byte CmpByte2=Flt->Type==FILTER_E8E9 ? 0xe9:0xe8;
-        for (uint CurPos=0;(int)CurPos<(int)DataSize-4;)
+        // DataSize is unsigned, so we use "CurPos+4" and not "DataSize-4"
+        // to avoid overflow for DataSize<4.
+        for (uint CurPos=0;CurPos+4<DataSize;)
         {
           byte CurByte=*(Data++);
           CurPos++;
@@ -424,7 +438,9 @@ byte* Unpack::ApplyFilter(byte *Data,uint DataSize,UnpackFilter *Flt)
     case FILTER_ARM:
       {
         uint FileOffset=(uint)WrittenFileSize;
-        for (uint CurPos=0;(int)CurPos<(int)DataSize-3;CurPos+=4)
+        // DataSize is unsigned, so we use "CurPos+3" and not "DataSize-3"
+        // to avoid overflow for DataSize<3.
+        for (uint CurPos=0;CurPos+3<DataSize;CurPos+=4)
         {
           byte *D=Data+CurPos;
           if (D[3]==0xeb) // BL command with '1110' (Always) condition.
@@ -440,6 +456,8 @@ byte* Unpack::ApplyFilter(byte *Data,uint DataSize,UnpackFilter *Flt)
       return SrcData;
     case FILTER_DELTA:
       {
+        // Unlike RAR3, we do not need to reject excessive channel
+        // values here, since RAR5 uses only 5 bits to store channel.
         uint Channels=Flt->Channels,SrcPos=0;
 
         FilterDstMemory.Alloc(DataSize);
@@ -503,6 +521,13 @@ void Unpack::UnpWriteData(byte *Data,size_t Size)
 }
 
 
+void Unpack::UnpInitData50(bool Solid)
+{
+  if (!Solid)
+    TablesRead5=false;
+}
+
+
 bool Unpack::ReadBlockHeader(BitInput &Inp,UnpackBlockHeader &Header)
 {
   Header.HeaderSize=0;
@@ -511,11 +536,11 @@ bool Unpack::ReadBlockHeader(BitInput &Inp,UnpackBlockHeader &Header)
     if (!UnpReadBuf())
       return false;
   Inp.faddbits((8-Inp.InBit)&7);
-  
+
   byte BlockFlags=Inp.fgetbits()>>8;
   Inp.faddbits(8);
   uint ByteCount=((BlockFlags>>3)&3)+1; // Block size byte count.
-  
+
   if (ByteCount==4)
     return false;
 
@@ -557,20 +582,20 @@ bool Unpack::ReadTables(BitInput &Inp,UnpackBlockHeader &Header,UnpackBlockTable
       return false;
 
   byte BitLength[BC];
-  for (int I=0;I<BC;I++)
+  for (uint I=0;I<BC;I++)
   {
-    int Length=(byte)(Inp.fgetbits() >> 12);
+    uint Length=(byte)(Inp.fgetbits() >> 12);
     Inp.faddbits(4);
     if (Length==15)
     {
-      int ZeroCount=(byte)(Inp.fgetbits() >> 12);
+      uint ZeroCount=(byte)(Inp.fgetbits() >> 12);
       Inp.faddbits(4);
       if (ZeroCount==0)
         BitLength[I]=15;
       else
       {
         ZeroCount+=2;
-        while (ZeroCount-- > 0 && I<sizeof(BitLength)/sizeof(BitLength[0]))
+        while (ZeroCount-- > 0 && I<ASIZE(BitLength))
           BitLength[I++]=0;
         I--;
       }
@@ -582,13 +607,13 @@ bool Unpack::ReadTables(BitInput &Inp,UnpackBlockHeader &Header,UnpackBlockTable
   MakeDecodeTables(BitLength,&Tables.BD,BC);
 
   byte Table[HUFF_TABLE_SIZE];
-  const int TableSize=HUFF_TABLE_SIZE;
-  for (int I=0;I<TableSize;)
+  const uint TableSize=HUFF_TABLE_SIZE;
+  for (uint I=0;I<TableSize;)
   {
     if (!Inp.ExternalBuffer && Inp.InAddr>ReadTop-5)
       if (!UnpReadBuf())
-        return(false);
-    int Number=DecodeNumber(Inp,&Tables.BD);
+        return false;
+    uint Number=DecodeNumber(Inp,&Tables.BD);
     if (Number<16)
     {
       Table[I]=Number;
@@ -597,7 +622,7 @@ bool Unpack::ReadTables(BitInput &Inp,UnpackBlockHeader &Header,UnpackBlockTable
     else
       if (Number<18)
       {
-        int N;
+        uint N;
         if (Number==16)
         {
           N=(Inp.fgetbits() >> 13)+3;
@@ -608,7 +633,16 @@ bool Unpack::ReadTables(BitInput &Inp,UnpackBlockHeader &Header,UnpackBlockTable
           N=(Inp.fgetbits() >> 9)+11;
           Inp.faddbits(7);
         }
-        if (I>0)
+        if (I==0)
+        {
+          // We cannot have "repeat previous" code at the first position.
+          // Multiple such codes would shift Inp position without changing I,
+          // which can lead to reading beyond of Inp boundary in mutithreading
+          // mode, where Inp.ExternalBuffer disables bounds check and we just
+          // reserve a lot of buffer space to not need such check normally.
+          return false;
+        }
+        else
           while (N-- > 0 && I<TableSize)
           {
             Table[I]=Table[I-1];
@@ -617,7 +651,7 @@ bool Unpack::ReadTables(BitInput &Inp,UnpackBlockHeader &Header,UnpackBlockTable
       }
       else
       {
-        int N;
+        uint N;
         if (Number==18)
         {
           N=(Inp.fgetbits() >> 13)+3;
@@ -632,17 +666,18 @@ bool Unpack::ReadTables(BitInput &Inp,UnpackBlockHeader &Header,UnpackBlockTable
           Table[I++]=0;
       }
   }
+  TablesRead5=true;
   if (!Inp.ExternalBuffer && Inp.InAddr>ReadTop)
-    return(false);
+    return false;
   MakeDecodeTables(&Table[0],&Tables.LD,NC);
   MakeDecodeTables(&Table[NC],&Tables.DD,DC);
   MakeDecodeTables(&Table[NC+DC],&Tables.LDD,LDC);
   MakeDecodeTables(&Table[NC+DC+LDC],&Tables.RD,RC);
-  return(true);
+  return true;
 }
 
 
 void Unpack::InitFilters()
 {
-  Filters.Reset();
+  Filters.SoftReset();
 }
